/*
 * Xbox input device driver for Linux - v0.1.4
 *
 * supports DVD playback kit IR receivers for Xbox
 *
 *	this is a very simple backend for LIRC, to be used
 *	together with the lirc_xir plugin for lirc_dev
 *
 * Copyright (c)  2004  Marko Friedemann <mfr@bmx-chemnitz.de>
 *
 * loosely affiliated to the usb-xboxir driver by
 *	Steven Toth <steve@toth.demon.co.uk>
 *
 * Released under GPL. See xpad-core.c for details
 */

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/module.h>
#include <linux/smp_lock.h>
#include <linux/usb.h>
#include <linux/version.h>

#define __USB_XPAD_IR
#include "xpad.h"
#undef __USB_XPAD_IR

#define XPAD_IR_RXBUF_LEN	   30
#define XPAD_IR_RXBUF_INC(var)  (var) += 1; (var) %= XPAD_IR_RXBUF_LEN

/* FIXME: avoid this static variable */
static struct usb_xpad *xpad_ir_xir = NULL;

extern struct xpad_device xpad_device[];

int xpad_ir_init(struct usb_xpad *xpad, int probedDevNum)
{
	xpad_ir_xir = xpad;
	
	/* initialize the ringbuffer */
	xpad_ir_xir->rx.buf = kmalloc(XPAD_IR_RXBUF_LEN, GFP_KERNEL);
	if (xpad_ir_xir->rx.buf == NULL) {
		err("cannot allocate memory for new IR input buffer");
		return -ENOMEM;
	}
	
	memset(xpad_ir_xir->rx.buf, 0, XPAD_IR_RXBUF_LEN);
	
	/* just set the input device name for the /proc output */
	xpad->dev.name = xpad_device[probedDevNum].name;
	
	info("%s", xpad->dev.name);
	return 0;
}

void xpad_ir_cleanup(void)
{
	kfree(xpad_ir_xir->rx.buf);
	xpad_ir_xir = NULL;
}

static void xpad_ir_process_packet(struct usb_xpad *xpad, u16 cmd, unsigned char *data)
{
/*	info("irpp: %02x %02x %02x %02x %02x %02x", data[0], data[1],
		data[2], data[3], data[4], data[5]);*/

	/* CHECKME: shouldn't be no sync neccessary here, should there? */
	if (1 > CIRC_SPACE(xpad_ir_xir->rx.head, xpad_ir_xir->rx.tail,
			   XPAD_IR_RXBUF_LEN))
	{
		info("no space left in buffer: keypress lost");
		return;
	}

	xpad_ir_xir->rx.buf[xpad_ir_xir->rx.tail] = data[2];
	XPAD_IR_RXBUF_INC(xpad_ir_xir->rx.tail);
}

void xpad_ir_irq_in(struct urb *urb)
{
	struct usb_xpad *xpad = urb->context;
	
	if (urb->status) {
		err("urb status");
		return;
	}
	
	xpad_ir_process_packet(xpad, 0, xpad->idata);
}

static unsigned char xpad_ir_get_keycode(void)
{
	unsigned char retval;

	/* CHECKME: shouldn't be no sync neccessary here, should there? */
	if (0 == CIRC_CNT(xpad_ir_xir->rx.head, xpad_ir_xir->rx.tail,
			  XPAD_IR_RXBUF_LEN))
		return 0;

	retval = (unsigned char)xpad_ir_xir->rx.buf[xpad_ir_xir->rx.head];
	XPAD_IR_RXBUF_INC(xpad_ir_xir->rx.head);
	
	return retval;
}

/**
 *	xpad_ir_lirc_open
 *
 *	Called from lirc_xir upon usage by lircd.
 */
static int xpad_ir_lirc_open(void)
{
	int status;
	
	if (xpad_ir_xir->open_count++)
		return 0;
		
	info("opening device");
	
	xpad_ir_xir->irq_in->dev = xpad_ir_xir->udev;
	if ((status = usb_submit_urb(xpad_ir_xir->irq_in))) {
		err("open input urb failed: %d", status);
		return -EIO;
	}

	return 0;
}

/**
 *	xpad_ir_lirc_close
 *
 *	Called from lirc_xir upon close event from lircd.
 */
static void xpad_ir_lirc_close(void)
{
	if (!--xpad_ir_xir->open_count) {
	
		info("closing device");
	
		usb_unlink_urb(xpad_ir_xir->irq_in);
	}
}

EXPORT_SYMBOL(xpad_ir_get_keycode);
EXPORT_SYMBOL(xpad_ir_lirc_open);
EXPORT_SYMBOL(xpad_ir_lirc_close);
