/*
 * Xbox input device driver for Linux - v0.1.4
 *
 *	rumble stuff
 *
 *	NOTE: this DOES NOT WORK without a patch to the force feedback code!
 *
 * Copyright (c)  2003, 2004  Marko Friedemann <mfr@bmx-chemnitz.de>
 *
 * Released under GPL. See xpad-core.c for details
 */

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/module.h>
#include <linux/smp_lock.h>
#include <linux/proc_fs.h>
#include <linux/usb.h>
#include <linux/version.h>
#include <asm/uaccess.h>

#define __USB_XPAD_RUMBLE
#include "xpad.h"
#undef __USB_XPAD_RUMBLE


static int rumble_on_load = 0;
MODULE_PARM( rumble_on_load, "i" );
MODULE_PARM_DESC( rumble_on_load, "set to 1 to activate ff support on insmod (default 0)" );


/*static*/ void xpad_rumble_proc_read_info(char *buf, char **start, off_t offset, int count, int *eof, struct usb_xpad *xpad, int *len)
{
	*len += sprintf( buf+*len, "rumble support available:\tyes\n" );
	*len += sprintf( buf+*len, "    automatically activated:\t%s\n\n",
		rumble_on_load ? "yes" : "no" );
}

/*static*/ void xpad_rumble_proc_read_unit_info(char *buf, char **start, off_t offset, int count, int *eof, struct usb_xpad *xpad, int *len)
{
	*len += sprintf( buf+*len, "rumble enabled:\t%s\n", xpad->rumble_enabled ? "yes" : "no" );
}

static int xpad_rumble_proc_read_feature_info(char *buf, char **start, off_t offset, int count, int *eof, void *data)
{
	int len = 0;
	struct usb_xpad *xpad = (struct usb_xpad *)data;
	
	len += sprintf( buf+len, "%d", xpad->rumble_enabled );
	*eof = 1;
	
	return len;
}

/*static*/ void xpad_rumble_proc_unit_create(struct usb_xpad *xpad, struct proc_dir_entry *entry)
{
	struct proc_dir_entry *tmp;
	
	/* world readable file */
	int flags = S_IFREG | S_IRUGO;

	tmp = create_proc_entry("rumble", flags, entry);
	tmp->data = xpad;
	tmp->read_proc = xpad_rumble_proc_read_feature_info;
}

/**
 *	xpad_rumble_send
 *
 *	Sets the output packet to control the rumble effects.
 */
static void xpad_rumble_send(struct usb_xpad *xpad)
{
//	DECLARE_WAITQUEUE(wait, current);
//	int timeout = HZ; /* 1 second */
	unsigned long flags;	/* holds the irq state from before the lock */
	int c;
	int reallen;
	int status;
	char data[6];
//	char *o;
		
//	set_current_state(TASK_INTERRUPTIBLE);
//	add_wait_queue(&xpad->wait, &wait);

	spin_lock_irqsave(&xpad->tx_lock, flags);
	
//	info("trying to send: %d %d", xpad->tx.head, xpad->tx.tail);
	
	// ring buffer empty -> leave running state
	if (xpad->tx.head == xpad->tx.tail) {
//		info("rumble packets sent, buffer empty");
		clear_bit(XPAD_TX_RUNNING, xpad->tx_flags);
		spin_unlock_irqrestore(&xpad->tx_lock, flags);
		return;
	}
	
	/* needed? don't think so, already set up by rumble_probe() 
	xpad->irq_out->transfer_buffer = xpad->tx_data;
	xpad->irq_out->transfer_buffer_length = XPAD_PKT_LEN_FF;
	xpad->irq_out->dev = xpad->udev;
	*/
	
	/* copy XPAD_PKT_LEN_FF bytes from ring into the transfer buffer
	 *  consider the ring buffer wraparound properly */
	c = min(XPAD_PKT_LEN_FF, CIRC_CNT_TO_END(
		xpad->tx.head, xpad->tx.tail, XPAD_TX_BUFSIZE));
	/*memcpy(xpad->irq_out->transfer_buffer,
		&xpad->tx.buf[xpad->tx.tail], c);
	if (c < XPAD_PKT_LEN_FF)
		memcpy(xpad->irq_out->transfer_buffer + c,
			&xpad->tx.buf[0], XPAD_PKT_LEN_FF - c);*/
	memcpy(data, &xpad->tx.buf[xpad->tx.tail], c);
	if (c < XPAD_PKT_LEN_FF)
		memcpy(data + c, &xpad->tx.buf[0], XPAD_PKT_LEN_FF - c);
	
	XPAD_TX_INC(xpad->tx.tail, XPAD_PKT_LEN_FF);
	
//	o = (unsigned char *) xpad->irq_out->transfer_buffer;
	
//	info("sending %02x %02x %02x %02x %02x %02x", o[0], o[1], o[2], o[3], o[4], o[5]);

	status = usb_bulk_msg(xpad->udev, usb_sndbulkpipe(xpad->udev, xpad->ep_out_adr),
		data, XPAD_PKT_LEN_FF, &reallen, 100);
	if (status != 0)
		err("usb_bulk_msg failed after sending %d bytes: %d", reallen, status);

	/*if (usb_submit_urb(xpad->irq_out)) {
		err("xpad_rumble: submit urb failed");
//		set_current_state(TASK_RUNNING);
//		remove_wait_queue(&xpad->wait, &wait);
//		return;
	}*/
	
	spin_unlock_irqrestore(&xpad->tx_lock, flags);
	
//	while (timeout && xpad->irq_out->status == -EINPROGRESS)
//		timeout = schedule_timeout(timeout);
	
//	set_current_state(TASK_RUNNING);
//	remove_wait_queue(&xpad->wait, &wait);
	
//	if (!timeout)
//		usb_unlink_urb(xpad->irq_out);
	
	xpad_rumble_send(xpad);
	return;
}

/**
 *	xpad_process_effect
 *
 *	Called via the dev entry to start/stop a rumble effect.
 */
static int xpad_process_effect(struct input_dev *dev, unsigned int type, unsigned int code, int value)
{
	struct usb_xpad *xpad = dev->private;
	int head, tail, empty;
	unsigned long flags;	/* holds the irq state from before the lock */
//	DECLARE_WAITQUEUE(wait, current);
	
	if (type != EV_FF)
		return -EINVAL;
	
	if (code >= 2)
		return -EINVAL;

	switch (code) {
		case 0:
			xpad->play_strong = !!value;
			break;
		case 1:
			xpad->play_weak = !!value;
			break;
		default:
			return -EINVAL;
	}
	
//	info("processing effect");
	
	spin_lock_irqsave(&xpad->tx_lock, flags);
	
	head = xpad->tx.head;
	tail = xpad->tx.tail;
	
	// get the free buffer space, need at leadt XPAD_PKT_LEN_FF (6) bytes
	if (CIRC_SPACE(head, tail, XPAD_TX_BUFSIZE) < XPAD_PKT_LEN_FF) {
		warn("not enough space in tx buffer to send new ff packet");
		spin_unlock_irqrestore(&xpad->tx_lock, flags);
		return -EINVAL;
	}
	empty = (head == tail);
	
	// set the buffers head XPAD_PKT_LEN_FF (6) bytes further
	XPAD_TX_INC(xpad->tx.head, XPAD_PKT_LEN_FF);
//	info("head before=%d after=%d", head, xpad->tx.head);
	
	// insert the data into the ringbuffer, increase internal head
	xpad->tx.buf[head] = 0;
	XPAD_TX_INC(head, 1);
	xpad->tx.buf[head] = XPAD_PKT_LEN_FF;
	XPAD_TX_INC(head, 1);
	xpad->tx.buf[head] = 0;
	XPAD_TX_INC(head, 1);
	xpad->tx.buf[head] = (xpad->play_strong ? xpad->strong_rumble : 0);
	XPAD_TX_INC(head, 1);
	xpad->tx.buf[head] = 0;
	XPAD_TX_INC(head, 1);
	xpad->tx.buf[head] = (xpad->play_weak ? xpad->weak_rumble : 0);
	XPAD_TX_INC(head, 1);
	
	spin_unlock_irqrestore(&xpad->tx, flags);
	
//	info("about to try and send: %d", empty);
	
	// buffer empty and no transfer running -> set running state && start
	if (empty && !test_and_set_bit(XPAD_TX_RUNNING, xpad->tx_flags)) {
//		info("starting to rumble");
		xpad_rumble_send(xpad);
	} /*else {
		info("not starting to rumble");
	}*/
	
//	set_current_state(TASK_INTERRUPTIBLE);
//	add_wait_queue(&xpad->wait, &wait);
	
	return 0;
}

static void xpad_rumble_irq_out(struct urb *urb)
{
	struct usb_xpad *xpad = urb->context;
	
//	info("urb out completed: %d bytes sent, status=%d", urb->actual_length, urb->status);

	if (waitqueue_active(&xpad->wait))
		wake_up(&xpad->wait);
	
	if (urb->status) {
		err("urb status: %d", urb->status);
		return;
	}
	
	xpad_rumble_send(xpad);
}

/**
 *	xpad_upload_effect
 *
 *	Called via ioctl on the dev entry to upload (setup) a rumble effect.
 */
static int xpad_upload_effect(struct input_dev *dev, struct ff_effect *effect)
{
	struct usb_xpad *xpad = dev->private;
	int i;
	int new_effect = -1;
	
	if (effect->type != FF_RUMBLE) {
		err( "can only handle rumble effects" );
		return -EINVAL;

	for (i=0; i<XPAD_FF_MAX_EFFECTS; ++i) {
		if (xpad->ff_effects[i] == NULL) {
			new_effect = i;
			break;
		}
	}
	if (new_effect == -1) {
		err( "cannot handle morr than %d effects", XPAD_FF_MAX_EFFECTS );
		return -EINVAL;
	}
	
	if ((xpad->ff_effects[new_effect] = kmalloc(
		sizeof(struct usb_xpad), GFP_KERNEL)) == NULL)
	{
		err( "not enought memory to create new effect" );
		return -ENOMEM;
	}
	memcpy(xpad->ff_effects[new_effect], effect, sizeof(struct ff_effect));
	
	return new_effect;
}

/**
 *	xpad_erase_effect
 *
 *	Called via ioctl on the dev entry to erase (remove) a rumble effect.
 */
static int xpad_erase_effect(struct input_dev *dev, int effect_id)
{
	struct usb_xpad *xpad = dev->private;

	/* FIXME: change this to something more advanced */
	switch (effect_id) {
		case 0:
			xpad->play_strong = 0;
			xpad->strong_rumble = 0;
			break;
		case 1:
			xpad->play_weak = 0;
			xpad->weak_rumble = 0;
			break;
		default:
			return -EINVAL;
	}
	return 0;
}

/*static*/ void xpad_rumble_open(struct usb_xpad *xpad)
{
	// setup the rumble stuff (FIXME: this whole rumble stuff is a mess)
	xpad->play_strong = 0;
	xpad->strong_rumble = 0;
	xpad->play_weak = 0;
	xpad->weak_rumble = 0;
}

/*static*/ void xpad_rumble_close(struct usb_xpad *xpad)
{
//	if (waitqueue_active(&xpad->wait))
//		wake_up(&xpad->wait);

	if (test_bit(XPAD_TX_RUNNING, &xpad->tx_flags))
		err("still sending rumble packets");

	// just to be sure, dunno if this is needed at all.
//	usb_unlink_urb(xpad->irq_out);
	
	xpad->tx.head = xpad->tx.tail = 0;
	
/*	switch (xpad->irq_out->status) {
		case 0:
		case -ECONNABORTED:
			/* those are OK, synch call results in 0,
			   asynch in CONNABORTED if success */
/*			break;
		case -ENOENT:
			err("output urb status: -ENOENT");
			break;
		case -EINPROGRESS:
			err("output urb status: -EINPROGRESS");
			break;
		case -EPROTO:
			err("output urb status: -EPROTO");
			break;
		case -EILSEQ:
			err("output urb status: -EILSEQ");
			break;
		case -ETIMEDOUT:
			err("output urb status: -ETIMEDOUT");
			break;
		case -ENOMEM:
			err("output urb status: -ENOMEM");
			break;
		case -ENODEV:
			err("output urb status: -ENODEV");
			break;
		case -ENXIO:
			err("output urb status: -ENXIO");
			break;
		case -EINVAL:
			err("output urb status: -EINVAL");
			break;
		case -EAGAIN:
			err("output urb status: -EAGAIN");
			break;
		case -EFBIG:
			err("output urb status: -EFBIG");
			break;
		case -EPIPE:
			err("output urb status: -EPIPE");
			break;
		case -EMSGSIZE:
			err("output urb status: -EMSGSIZE");
			break;
		default:
			err("output urb status: unknown error %d", xpad->irq_out->status);
	}*/
}

/**
 *	xpad_ioctl
 *
 *	Called via ioctl to change driver parameters (toggle mouse/ff support).
 *	Note that this relies on some unofficial changes to the input subsystem.
 */ 
/*static*/ int xpad_rumble_ioctl(struct input_dev *dev, unsigned int cmd, unsigned long arg)
{
	struct usb_xpad *xpad = (struct usb_xpad *) dev->private;

	switch (cmd) {
		case USB_XPAD_IOCSRUMBLE:
			return get_user( xpad->rumble_enabled, (int *) arg );
		case USB_XPAD_IOCGRUMBLE:
			return put_user( xpad->rumble_enabled, (int *) arg );
		default:
			return -ENOTTY;
	}
}

/*static*/ int xpad_rumble_probe(struct usb_device *udev, struct usb_xpad *xpad, unsigned int ifnum)
{
	struct usb_endpoint_descriptor *ep_out;

	/* setup output interrupt pipe (rumble control) */
	/*xpad->irq_out = usb_alloc_urb(0);
        if (!xpad->irq_out) {
		err("cannot allocate memory for irq out urb");
                return -1;
        }*/
	
	ep_out = udev->actconfig->interface[ifnum].altsetting[0].endpoint + 1;
	xpad->ep_out_adr = ep_out->bEndpointAddress;
	// URB length defined with XPAD_PKT_LEN_FF (==6 or MS pads will choke)
	/*FILL_INT_URB(xpad->irq_out, udev,
		usb_sndintpipe(udev, ep_irq_out->bEndpointAddress),
		xpad->tx_data, XPAD_PKT_LEN_FF, xpad_rumble_irq_out, xpad, 0);*/

	xpad->dev.event = xpad_process_effect;
	xpad->dev.upload_effect = xpad_upload_effect;
	xpad->dev.erase_effect = xpad_erase_effect;
	xpad->rumble_enabled = rumble_on_load;
	
	/* rumble output */
	set_bit(EV_FF, xpad->dev.evbit);
	set_bit(FF_RUMBLE, xpad->dev.ffbit);
	xpad->dev.ff_effects_max = 2;
	
	init_waitqueue_head(&xpad->wait);
	
	spin_lock_init(&xpad->tx_lock);
	xpad->tx.buf = xpad->tx_buf;
	
	return 0;
}

/*static*/ void xpad_rumble_disconnect(struct usb_xpad *xpad)
{
	//usb_free_urb(xpad->irq_out);
}
